#include "interfaces/idapro/all.h"
// #include "interfaces/STARSTypes.h"
#include "base/SMPDataFlowAnalysis.h"
#include "interfaces/SMPDBInterface.h"
#include "interfaces/abstract/all.h"

#if 0
#include <pro.h>
#include <ua.hpp>
#include <intel.hpp>
#endif

using namespace std;

// Constructors
STARS_IDA_op_t::STARS_IDA_op_t(op_t IDAOp) : m_Opnd(IDAOp) {
#if (IDA_SDK_VERSION < 700)
	if ((this->m_Opnd.dtyp == ((char)-1)) && (this->m_Opnd.type != o_imm)) {
		this->m_Opnd.dtyp = global_STARS_program->GetSTARS_ISA_dtyp();
	}
#else
	if ((this->m_Opnd.dtype == ((char)-1)) && (this->m_Opnd.type != o_imm)) {
		this->m_Opnd.dtype = global_STARS_program->GetSTARS_ISA_dtyp();
	}
#endif
};

// Clone method (deep copy).
STARSOpndTypePtr STARS_IDA_op_t::clone(void) const {
	auto TempOpPtr(std::make_shared<STARS_IDA_op_t>(this->m_Opnd));
	return std::dynamic_pointer_cast<STARS_op_t>(TempOpPtr);
} // end of STARS_IDA_op_t::clone()

// Less-than operator for use in STL ordered containers, e.g. sets.
bool STARS_IDA_op_t::operator<(const STARS_op_t &rOp) const {
	unsigned char Type1 = this->GetOpType();
	unsigned char Type2 = rOp.GetOpType();
	if (Type1 != Type2)
		return (Type1 < Type2);
	switch (Type1) {
		case o_void: return false;
		case o_reg: return MDLessReg(this->GetReg(), rOp.GetReg());
		case o_mem: return ((this->GetAddr() < rOp.GetAddr())
			|| (this->HasSIBByte() && rOp.HasSIBByte() && (this->GetAddr() == rOp.GetAddr()) && (this->MDGetSIBIndexReg() < rOp.MDGetSIBIndexReg())));
		case o_phrase: if (this->HasSIBByte() && rOp.HasSIBByte()) return ((this->GetSIB() < rOp.GetSIB()) || ((this->GetSIB() == rOp.GetSIB()) && (this->GetSpecFlag4() < rOp.GetSpecFlag4())));
						else if (rOp.HasSIBByte()) return true;  // no SIB < has SIB
						else if (this->HasSIBByte()) return false; // no SIB < has SIB
						else return MDLessReg(this->GetReg(), rOp.GetReg()); // no SIB bytes
		case o_displ: if (this->HasSIBByte() && rOp.HasSIBByte())
						  return ((this->GetSIB() < rOp.GetSIB()) 
							|| ((this->GetSIB() == rOp.GetSIB()) 
								&& ((this->GetAddr() < rOp.GetAddr()) || ((this->GetAddr() == rOp.GetAddr()) && (this->GetSpecFlag4() < rOp.GetSpecFlag4())))));
						else if (rOp.HasSIBByte()) return true;  // no SIB < has SIB
						else if (this->HasSIBByte()) return false; // no SIB < has SIB
						else return ((this->GetAddr() < rOp.GetAddr())
							|| ((this->GetAddr() == rOp.GetAddr()) && MDLessReg(this->GetReg(), rOp.GetReg()))); // no SIB bytes
		case o_imm: return (this->GetImmedValue() < rOp.GetImmedValue());
		case o_far:  // fall through to o_near case
		case o_near: return (this->GetAddr() < rOp.GetAddr());
		case o_trreg:  // fall through
		case o_dbreg:  // fall through
		case o_crreg:  // fall through
		case o_fpreg:  // fall through
		case o_mmxreg: // fall through
		case o_xmmreg: // fall through
		case o_ymmreg: return (this->GetReg() < rOp.GetReg()); // no subword regs to deal with, don't need MDLessReg()

		default: SMP_msg("ERROR: Unknown operand type in LessOp.\n"); return false;
	}; // end switch (Opnd1.type)
}; // end of operator less-than

// Get STARS SSA name index.
unsigned int STARS_IDA_op_t::GetOpGlobalIndex(void) const {
	unsigned int index = 0;
	index |= (((unsigned int) this->m_Opnd.offo) & 0x000000ff);
	index <<= 8;
	index |= (((unsigned int) this->m_Opnd.offb) & 0x000000ff);
	index <<= 8;
	index |= (((unsigned int) this->m_Opnd.n) & 0x000000ff);
	return index;
}

bool STARS_IDA_op_t::MD_STARS_op256(void) const {       // is VEX.L set?
	return ((this->GetSpecFlag4() & STARS_VEXPR) != 0) && ((this->GetSpecFlag4() & STARS_VEX_L) != 0);
}

bool STARS_IDA_op_t::MD_STARS_is_vsib(void) const {  // does instruction use VSIB variant of the sib byte?
	return ((this->GetSpecFlag4() & STARS_VSIB) != 0);
}

// Get x86 SIB byte base register
int STARS_IDA_op_t::MDGetSIBBaseReg(void) const {
	int base = this->GetSIB() & 7;
#ifdef __EA64__
	if (this->GetSpecFlag4() & STARS_REX_B)
		base |= 8;   // access new regs R8-R15
#endif
	return base;
}

// Get x86 SIB byte index register
short STARS_IDA_op_t::MDGetSIBIndexReg(void) const {
	short index = (short) ((this->GetSIB() >> 3) & 7);
#ifdef __EA64__
	if (this->GetSpecFlag4() & STARS_REX_X)
		index |= 8;
#endif
	if (this->MD_STARS_is_vsib())
		index += this->MD_STARS_op256() ? 81 /*R_ymm0*/ : 64 /*R_xmm0*/;
	return index;
}

uint16_t STARS_IDA_op_t::GetByteWidth(void) const {
	uint16_t DataSize;
	char OpDtyp = this->GetOpDtyp();
	if (this->IsRegOp()) {
		DataSize = (uint16_t) RegSizes[this->GetReg()];
		if (OpDtyp == dt_byte) {
			DataSize = 1;
		}
		else if (OpDtyp == dt_word) {
			DataSize = 2;
		}
		else if (OpDtyp == STARS_dt_dword) {
			DataSize = 4;
		}
		else if (OpDtyp == dt_qword) {
			DataSize = 8;
		}
		return DataSize;
	}

	switch (OpDtyp) {
	case dt_byte:
		DataSize = 1;
		break;
	case dt_word:
		DataSize = 2;
		break;
	case dt_dword:
	case dt_float:
	case dt_code:
	case dt_unicode:
	case dt_string:
		DataSize = 4;
		break;
	case dt_double:
	case dt_qword:
		DataSize = 8;
		break;
	case dt_tbyte:
		DataSize = 10;
		break;
	case dt_packreal:
		DataSize = 12;
		break;
	case dt_byte16:
	case dt_ldbl:
		DataSize = 16;
		break;
	case dt_fword:
		DataSize = 6;
		break;
#if (IDA_SDK_VERSION < 700)
	case dt_3byte:
		DataSize = 3;
		break;
#endif
	case dt_byte32:
		DataSize = 32;
		break;
	case dt_byte64:
		DataSize = 64;
		break;
	default:
		SMP_msg("ERROR: unexpected data type %d in STARS_IDA_op_t::GetByteWidth() \n", OpDtyp);
		DataSize = (uint16_t) global_STARS_program->GetSTARS_ISA_dtyp();
		break;
	}
	return DataSize;
} // end of STARS_IDA_op_t::GetByteWidth()

// Change default byte width of opnd
void STARS_IDA_op_t::SetByteWidth(uint16_t ByteWidth) {
#if (IDA_SDK_VERSION < 700)
	if (1 == ByteWidth) {
		this->m_Opnd.dtyp = dt_byte;
	}
	else if (2 == ByteWidth) {
		this->m_Opnd.dtyp = dt_word;
	}
	else if (4 == ByteWidth) {
		this->m_Opnd.dtyp = dt_dword;
	}
	else if (8 == ByteWidth) {
		this->m_Opnd.dtyp = dt_qword;
	}
#else
	if (1 == ByteWidth) {
		this->m_Opnd.dtype = dt_byte;
	}
	else if (2 == ByteWidth) {
		this->m_Opnd.dtype = dt_word;
	}
	else if (4 == ByteWidth) {
		this->m_Opnd.dtype = dt_dword;
	}
	else if (8 == ByteWidth) {
		this->m_Opnd.dtype = dt_qword;
	}
	else if (16 == ByteWidth) {
		this->m_Opnd.dtype = dt_byte16;
	}
#endif
	else {
		SMP_msg("ERROR: ByteWidth of %d specified for operandr\n", ByteWidth);
	}
	return;
}; // end of STARS_IDA_op_t::SetByteWidth()

void STARS_IDA_op_t::DoubleRegWidth(void) {
#if (IDA_SDK_VERSION < 700)
	if (this->m_Opnd.dtyp == dt_byte) {  // current width 1, e.g. register AL
		this->SetByteWidth(2);  // double the byte width
		// Transform byte reg into word reg; the dtyp field will determine the interpretation
		//  of the word reg as AX, EAX, or RAX, so only the AL case needs a register number change.
		this->m_Opnd.reg = (uint16_t) MDCanonicalizeSubReg((STARS_regnum_t) this->m_Opnd.reg);
	}
	else if (this->m_Opnd.dtyp == dt_word) {  // current width 2, e.g. register AX
		this->SetByteWidth(4);  // double the byte width; makes R_ax be interpreted as EAX, for example
	}
	else if (this->m_Opnd.dtyp == dt_dword) {  // current width 4, e.g. register EAX
		this->SetByteWidth(8);  // double the byte width; makes R_ax be interpreted as RAX, for example
	}
#else
	if (this->m_Opnd.dtype == dt_byte) {  // current width 1, e.g. register AL
		this->SetByteWidth(2);  // double the byte width
		// Transform byte reg into word reg; the dtype field will determine the interpretation
		//  of the word reg as AX, EAX, or RAX, so only the AL case needs a register number change.
		this->m_Opnd.reg = (uint16_t)MDCanonicalizeSubReg((STARS_regnum_t) this->m_Opnd.reg);
	}
	else if (this->m_Opnd.dtype == dt_word) {  // current width 2, e.g. register AX
		this->SetByteWidth(4);  // double the byte width; makes R_ax be interpreted as EAX, for example
	}
	else if (this->m_Opnd.dtype == dt_dword) {  // current width 4, e.g. register EAX
		this->SetByteWidth(8);  // double the byte width; makes R_ax be interpreted as RAX, for example
	}
#endif
	return;
} // end of STARS_IDA_op_t::DoubleRegWidth()

// Set STARS SSA name index into three unused bytes of the op_t structure.
void STARS_IDA_op_t::SetOpGlobalIndex(std::size_t index) {
	this->m_Opnd.n = (char)(index & 0x000000ff);
	this->m_Opnd.offb = (char)((index & 0x0000ff00) >> 8);
	this->m_Opnd.offo = (char)((index & 0x00ff0000) >> 16);
	return;
}

void STARS_IDA_op_t::SetReg(STARS_regnum_t NewReg) { 
	this->m_Opnd.reg = NewReg; 
	if (!this->IsMemOp()) {
		this->SetByteWidth(GetRegSize(NewReg));
	}
};

